package hondana.action;

import hondana.businesslogic.BusinessLogic;
import hondana.exception.BusinessException;
import hondana.exception.SystemException;
import hondana.view.View;
import hondana.object.Request;
import hondana.object.Session;
import hondana.object.ViewStack;
import hondana.object.impl.RequestImpl;
import hondana.object.impl.SessionImpl;
import hondana.object.impl.ViewStackImpl;
import hondana.util.AppActionMapping;
import hondana.util.Config;
import hondana.util.database.impl.QueryRunnerImpl;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.struts.action.Action;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;

/**
 * コントローラアクション
 */
public class ControllerAction extends Action {

    private Log logger = LogFactory.getLog(this.getClass());

    /////////////////////////////////////////////////////////////////
    // business logic method

    /**
     * 業務処理を行う。
     */
    public ActionForward execute(ActionMapping mapping, ActionForm form,
            HttpServletRequest servletRequest,
            HttpServletResponse servletReqponse) throws Exception {

        //====================
        // オブジェクトの取得
        //====================

        ServletContext context = getServlet().getServletContext();
        Config config = (Config) context.getAttribute(Config.CONFIG_KEY);
        Request request = new RequestImpl(servletRequest);
        Session session = new SessionImpl(servletRequest);
        AppActionMapping appmapping = (AppActionMapping) mapping;
        View in = (View) form;
        ViewStack stack = getViewStack(session);

        logger.debug("input view: " + in.getName());

        // 入力値をトリム。
        in.trim();

        //====================
        // ログイン状態の検査
        //====================

        if (appmapping.isValidateLogin()) {
            logger.debug("validate login");
            if (session.getUser() == null)
                throw new SystemException("access denied");
        }

        //================
        // トークンの検査
        //================

        if (appmapping.isValidateToken()) {
            logger.debug("validate token");
            if (!isTokenValid(servletRequest, true))
                throw new SystemException("not correct token");
        }

        //==================
        // アクションの取得
        //==================

        // アクション名を取得。
        String acname = appmapping.getAcname();
        if (acname == null)
            acname = appmapping.findAcname(in.getBtn());

        logger.debug("action: " + acname);

        // アクションを取得。
        Object action = getAction(acname);

        //====================
        // 業務ロジックの実行
        //====================

        View out = null;

        if (action instanceof BusinessLogic) {
            // アクションが業務ロジックの場合、業務ロジックを実行。
            BusinessLogic bl = (BusinessLogic) action;
            out = (View) executeBusinessLogic(bl, request, session, stack, in,
                    config);

            // 業務ロジックの実行に失敗した場合、トークンを再保存。
            if (appmapping.isValidateToken() && bl.isFailed()) {
                logger.debug("resave token");
                saveToken(servletRequest);
            }
        } else if (action instanceof View) {
            // アクションがビューの場合、ビューをそのまま出力。
            out = (View) action;
        } else {
            throw new SystemException("not correct action '" + acname + "'");
        }

        //================
        // トークンの保存
        //================

        if (appmapping.isSaveToken()) {
            logger.debug("save token");
            saveToken(servletRequest);
        }

        //========================
        // ビューをスタックに積む
        //========================

        if (appmapping.isStacked()) {
            // 「戻る」ではなければ、ビューをスタックに積む。。
            if (!out.isReturned()) {
                logger.debug("stack view: " + out.getName());
                stack.push(in);
            }
        } else {
            // フラグがセットされていない場合、スタックをクリア。
            stack.clear();
            logger.debug("clear view stack");
        }

        logger.debug("output view: " + out.getName());

        // リクエストにビューをセット。
        setView(request, out);

        // ビューに紐付くアクションフォワードを返す。
        return mapping.findForward(out.getName());
    }

    /**
     * 業務ロジックを実行する。
     */
    private Object executeBusinessLogic(BusinessLogic bl, Request request,
            Session session, ViewStack stack, View in, Config config)
            throws Exception {

        //====================
        // オブジェクトの注入
        //====================

        bl.setRequest(request);
        bl.setSession(session);
        bl.setViewStack(stack);
        bl.setView(in);

        //========================
        // クエリーランナーの生成
        //========================

        QueryRunnerImpl qr = null;
        boolean hasQR = bl.isNecessityQueryRunner();

        if (hasQR) {
            // クエリーランナーが必要な場合、クエリーランナーを注入。
            String url = config.getJdbcUrl();
            String user = config.getJdbcUser();
            String password = config.getJdbcPassword();
            String driver = config.getJdbcDriver();
            QueryRunnerImpl.registerDriver(driver);
            qr = new QueryRunnerImpl(url, user, password);
            bl.setQueryRunner(qr);
        }

        //====================
        // 業務ロジックの実行
        //====================

        Object out = null;

        try {
            out = bl.execute();

            // クエリーランナーを生成した場合、トランザクションをコミット。
            if (hasQR)
                qr.commit();
        } catch (BusinessException e) {
            // 業務エラーの処理
            logger.warn("business error: " + e.getCode());

            // クエリーランナーを生成した場合、トランザクションをロールバック。
            if (hasQR)
                qr.rollback();

            bl.setFailed(true);
            out = bl.error(e);

            // エラーコードをリクエストにセーブ。
            request.saveError(e.getCode());

            if (e.getCause() != null)
                logger.warn("business error cause", e.getCause());

        } catch (Exception e) {
            // 致命的なエラーの処理
            // クエリーランナーを生成した場合、トランザクションをロールバック。
            if (hasQR)
                qr.rollback();
            throw e;
        } finally {
            // クエリーランナーを生成した場合、コネクションをクローズ。
            if (hasQR)
                qr.close();
        }

        // 業務ロジックの戻り値が業務ロジックの場合、業務ロジックを再帰的に実行。
        if (out instanceof BusinessLogic)
            out = executeBusinessLogic((BusinessLogic) out, request, session,
                    stack, in, config);

        // ビューを返す。
        return out;
    }

    /////////////////////////////////////////////////////////////////
    // private method

    /**
     * アクションを取得する。
     */
    private Object getAction(String action) {
        try {
            // クラス名からインスタンスを生成して返す。
            return Class.forName(action).newInstance();
        } catch (InstantiationException e) {
            throw new SystemException(e);
        } catch (IllegalAccessException e) {
            throw new SystemException(e);
        } catch (ClassNotFoundException e) {
            throw new SystemException(e);
        }
    }

    /**
     * ビューをリクエストにセットする。
     */
    private void setView(Request request, View view) {
        // ビュー名をキーとして、ビューをリクエストにセット。
        request.set(view.getName(), view);
    }

    /**
     * ビュースタックを取得する。
     */
    private ViewStack getViewStack(Session session) {
        // セッションからビュースタックを取得する。
        ViewStack stack = session.getViewStack();

        // ビュースタックを取得できなかった場合、新しく生成。
        if (stack == null) {
            stack = new ViewStackImpl(10);
            session.setViewStack(stack);
        }

        // ビュースタックを返す。
        return stack;
    }

}